from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Optional
from typing import Union

from pydantic import Field

from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.enums.watchlists import WatchlistType


class Date(ResponseModel):
    day: Optional[int] = Field(
        None,
        description="Day of month. Must be from 1 to 31 and valid for the year and month, or 0\n if specifying a year by itself or a year and month where the day is not\n significant.",
    )
    month: Optional[int] = Field(
        None,
        description="Month of year. Must be from 1 to 12, or 0 if specifying a year without a\n month and day.",
    )
    year: Optional[int] = Field(
        None,
        description="Year of date. Must be from 1 to 9999, or 0 if specifying a date without\n a year.",
    )


class IncludedDepartment(ResponseModel):
    """
    A model representing a department included on a watchlist.

    **Fields**:

    * **added_time**: `datetime` - The time the department was included on the watchlist.
    * **name**: `str` - Department name. Example: "Engineering".
    """

    added_time: datetime = Field(None, alias="addedTime")
    name: Optional[str] = Field(None, examples=["Engineering"])


class IncludedDirectoryGroup(ResponseModel):
    """
    A model representing a directory group included on a watchlist.

    **Fields**:

    * **added_time**: `datetime` - The time the directory group was included on the watchlist.
    * **group_id**: `str` - A unique group ID for the directory group.
    * **is_deleted**: `bool` - Whether the included group was deleted by the directory provider but still referenced by the watchlist
    * **name**: `str` - The name of directory group. Example: "Research and Development".
    """

    added_time: datetime = Field(None, alias="addedTime")
    group_id: Optional[str] = Field(
        None, description="A unique group ID.", examples=["23"], alias="groupId"
    )
    is_deleted: Optional[bool] = Field(
        None,
        description="Whether the included group was deleted by the directory provider but still referenced by the watchlist",
        alias="isDeleted",
    )
    name: Optional[str] = Field(None, examples=["Research and development"])


class WatchlistUser(ResponseModel):
    """
    A model representing a user whose associated with a watchlist.

    **Fields**:

    * **added_time**: `datetime` - The time the user was associated with the watchlist.
    * **user_id**: `str` - Unique user ID.
    * **username**: `str - Username.
    """

    added_time: datetime = Field(None, alias="addedTime")
    user_id: Optional[str] = Field(
        None, description="A unique user ID.", examples=["23"], alias="userId"
    )
    username: Optional[str] = Field(None, examples=["foo@bar.com"])


class WatchlistActor(ResponseModel):
    """
    A model representing a user whose associated with a watchlist.

    **Fields**:

    * **added_time**: `datetime` - The time the user was associated with the watchlist.
    * **actor_id**: `str` - Unique actor ID.
    * **actor_name**: `str - Actor name.
    """

    added_time: datetime = Field(None, alias="addedTime")
    actor_id: Optional[str] = Field(
        None, description="A unique actor ID.", examples=["23"], alias="actorId"
    )
    actor_name: Optional[str] = Field(None, examples=["foo@bar.com"], alias="actorname")


class ExcludedUsersList(ResponseModel):
    """
    A model representing a list of users excluded from a watchlist.
    Excluded users are those that have been individually excluded from that list.

    **Fields**:

    * **excluded_users**: `List[WatchlistUser]` - The list of excluded users.
    * **total_count**: `int`
    """

    excluded_users: Optional[List[WatchlistUser]] = Field(None, alias="excludedUsers")
    total_count: Optional[int] = Field(
        None,
        description="The total count of all excluded users.",
        examples=[10],
        alias="totalCount",
    )


class ExcludedActorsList(ResponseModel):
    """
    A model representing a list of actors excluded from a watchlist.
    Excluded actors are those that have been individually excluded from that list.

    **Fields**:

    * **excluded_actors**: `List[WatchlistActor]` - The list of excluded actors.
    * **total_count**: `int`
    """

    excluded_actors: Optional[List[WatchlistActor]] = Field(
        None, alias="excludedActors"
    )
    total_count: Optional[int] = Field(
        None,
        description="The total count of all excluded actors.",
        examples=[10],
        alias="totalCount",
    )


class IncludedDepartmentsList(ResponseModel):
    """
    A model representing a list of departments included on a watchlist.

    **Fields**:

    * **included_departments**: `List[IncludedDepartment]` - The list of included departments.
    * **total_count**: `int` - The total count of all included departments.
    """

    included_departments: Optional[List[IncludedDepartment]] = Field(
        None, alias="includedDepartments"
    )
    total_count: Optional[int] = Field(
        None,
        description="The total count of all included departments.",
        examples=[10],
        alias="totalCount",
    )


class IncludedDirectoryGroupsList(ResponseModel):
    """
    A model representing a list of directory groups included on a watchlist.

    **Fields**:

    * **included_directory_groups**: `List[IncludedDirectoryGroup]` - The list of included directory groups.
    * **total_count**: `int` - The total count of all included directory groups.
    """

    included_directory_groups: Optional[List[IncludedDirectoryGroup]] = Field(
        None, alias="includedDirectoryGroups"
    )
    total_count: Optional[int] = Field(
        None,
        description="The total count of all included directory groups.",
        examples=[10],
        alias="totalCount",
    )


class IncludedUsersList(ResponseModel):
    """
    A model representing a list of users included on a watchlist.
    Included users are those that have been individually included on that list.

    * **included_users**: `List[WatchlistUser]` - The list of included users.
    * **total_count**: `int` - The total count of all included users.
    """

    included_users: Optional[List[WatchlistUser]] = Field(None, alias="includedUsers")
    total_count: Optional[int] = Field(
        None,
        description="The total count of all included users.",
        examples=[10],
        alias="totalCount",
    )


class IncludedActorsList(ResponseModel):
    """
    A model representing a list of actors included on a watchlist.
    Included users are those that have been individually included on that list.

    * **included_actors**: `List[WatchlistActor]` - The list of included actors.
    * **total_count**: `int` - The total count of all included actors.
    """

    included_actors: Optional[List[WatchlistActor]] = Field(
        None, alias="includedActors"
    )
    total_count: Optional[int] = Field(
        None,
        description="The total count of all included actors.",
        examples=[10],
        alias="totalCount",
    )


class WatchlistStatsV2(ResponseModel):
    """
    A model representing stats for a watchlist.

    **Fields**:

    * **included_actors_count**: `int` - The number of actors explicitly included on the watchlist.
    * **included_departments_count**: `int` - The number of departments explicitly included on the watchlist.
    * **included_directory_groups_count**: `int` - The number of directory groups explicitly included on the watchlist.
    * **excluded_actors_count**: `int` - The number of actors explicitly excluded from the watchlist.
    * **excluded_departments_count**: `int` - The number of departments explicitly excluded from the watchlist.
    * **excluded_directory_groups_count**: `int` - The number of directory groups explicitly excluded from the watchlist.
    """

    included_departments_count: Optional[int] = Field(
        None,
        description="The number of departments explicitly included on the watchlist.",
        alias="includedDepartmentsCount",
        table=lambda included_departments_count: included_departments_count or 0,
    )
    included_directory_groups_count: Optional[int] = Field(
        None,
        description="The number of directory groups explicitly included on the watchlist.",
        alias="includedDirectoryGroupsCount",
        table=lambda included_directory_groups_count: included_directory_groups_count
        or 0,
    )
    included_actors_count: Optional[int] = Field(
        None,
        description="The number of actors explicitly included on the watchlist.",
        alias="includedActorsCount",
        table=lambda included_users_count: included_users_count or 0,
    )
    excluded_actors_count: Optional[int] = Field(
        None,
        description="The number of actors explicitly excluded from the watchlist.",
        alias="excludedActorsCount",
        # displays a value of None as 0
        table=lambda excluded_users_count: excluded_users_count or 0,
    )
    excluded_departments_count: Optional[int] = Field(
        None,
        description="The number of departments explicitly excluded from the watchlist.",
        alias="excludedDepartmentsCount",
        table=lambda included_departments_count: included_departments_count or 0,
    )
    excluded_directory_groups_count: Optional[int] = Field(
        None,
        description="The number of directory groups explicitly excluded from the watchlist.",
        alias="excludedDirectoryGroupsCount",
        table=lambda included_directory_groups_count: included_directory_groups_count
        or 0,
    )


class WatchlistStats(ResponseModel):
    """
    A model representing stats for a watchlist.

    **Fields**:

    * **excluded_users_count**: `int` - The number of users explicitly excluded from the watchlist.
    * **included_departments_count**: `int` - The number of departments explicitly included on the watchlist.
    * **included_directory_groups_count**: `int` - The number of directory groups explicitly included on the watchlist.
    * **included_users_count**: `int` - The number of users explicitly included on the watchlist.
    """

    excluded_users_count: Optional[int] = Field(
        None,
        description="The number of users explicitly excluded from the watchlist.",
        alias="excludedUsersCount",
        # displays a value of None as 0
        table=lambda excluded_users_count: excluded_users_count or 0,
    )
    included_departments_count: Optional[int] = Field(
        None,
        description="The number of departments explicitly included on the watchlist.",
        alias="includedDepartmentsCount",
        table=lambda included_departments_count: included_departments_count or 0,
    )
    included_directory_groups_count: Optional[int] = Field(
        None,
        description="The number of directory groups explicitly included on the watchlist.",
        alias="includedDirectoryGroupsCount",
        table=lambda included_directory_groups_count: included_directory_groups_count
        or 0,
    )
    included_users_count: Optional[int] = Field(
        None,
        description="The number of users explicitly included on the watchlist.",
        alias="includedUsersCount",
        table=lambda included_users_count: included_users_count or 0,
    )


class WatchlistMembersList(ResponseModel):
    """
    A model representing a list of watchlist members.
    Watchlist members are users who are on a list, whether it is because they are individually included,
    or because they are part of a department or directory group that is included.


    **Fields**:

    * **watchlist_members**: `List[WatchlistUser]` - The list of watchlist members.
    * **total_count**: `int` - Total count of members on the watchlist.
    """

    total_count: Optional[int] = Field(
        None,
        description="The total count of all included users.",
        examples=[10],
        alias="totalCount",
    )
    watchlist_members: Optional[List[WatchlistUser]] = Field(
        None, alias="watchlistMembers"
    )


class WatchlistMembersListV2(ResponseModel):
    """
    A model representing a list of watchlist members.
    Watchlist members are users who are on a list, whether it is because they are individually included,
    or because they are part of a department or directory group that is included.


    **Fields**:

    * **watchlist_members**: `List[WatchlistActor]` - The list of watchlist members.
    * **total_count**: `int` - Total count of members on the watchlist.
    """

    total_count: Optional[int] = Field(
        None,
        description="The total count of all included actors.",
        examples=[10],
        alias="totalCount",
    )
    watchlist_members: Optional[List[WatchlistActor]] = Field(
        None, alias="watchlistMembers"
    )


class Watchlist(ResponseModel):
    """
    A model representing an Incydr Watchlist.

    **Fields**:

    * **description**: `str` - Optional description for a custom watchlist.
    * **list_type**: [`WatchlistType`][watchlist-types] - The watchlist type.
    * **stats**: `WatchlistStats` - Watchlist membership information.  Includes `included_user_count`, `included_department_count`, `included_directory_groups_count`, and `excluded_users_count`.
    * **tenant_id**: `str` - A unique tenant ID.
    * **title**: `str` - Title for a custom watchlist.
    * **watchlist_id**: `str` - A unique watchlist ID.
    """

    description: Optional[str] = Field(
        None, description="Description for a custom watchlist."
    )
    list_type: Union[WatchlistType, str] = Field(alias="listType")
    stats: Optional[WatchlistStats] = None
    tenant_id: Optional[str] = Field(
        None, description="A unique tenant ID.", alias="tenantId"
    )
    title: Optional[str] = Field(None, description="Title for a custom watchlist.")
    watchlist_id: Optional[str] = Field(
        None, description="A unique watchlist ID.", alias="watchlistId"
    )


class WatchlistV2(ResponseModel):
    """
    A model representing an Incydr Watchlist.

    **Fields**:

    * **description**: `str` - Optional description for a custom watchlist.
    * **list_type**: [`WatchlistType`][watchlist-types] - The watchlist type.
    * **stats**: `WatchlistStatsV2` - Watchlist membership information.
    * **tenant_id**: `str` - A unique tenant ID.
    * **title**: `str` - Title for a custom watchlist.
    * **watchlist_id**: `str` - A unique watchlist ID.
    """

    description: Optional[str] = Field(
        None, description="Description for a custom watchlist."
    )
    list_type: Union[WatchlistType, str] = Field(alias="listType")
    stats: Optional[WatchlistStatsV2] = None
    tenant_id: Optional[str] = Field(
        None, description="A unique tenant ID.", alias="tenantId"
    )
    title: Optional[str] = Field(None, description="Title for a custom watchlist.")
    watchlist_id: Optional[str] = Field(
        None, description="A unique watchlist ID.", alias="watchlistId"
    )


class WatchlistsPage(ResponseModel):
    """
    A model representing a page of `Watchlist` objects.

    **Fields**:

    * **total_count**: `int` - Total count of watchlists found by the query.
    * **watchlists**: `List[Watchlist]` - The list `n` number of watchlists retrieved from the query, where `n=page_size`.
    """

    total_count: Optional[int] = Field(
        None,
        description="The total count of all watchlists.",
        examples=[10],
        alias="totalCount",
    )
    watchlists: Optional[List[Watchlist]] = Field(
        None, description="The list of watchlists."
    )


class WatchlistsPageV2(ResponseModel):
    """
    A model representing a page of `Watchlist` objects.

    **Fields**:

    * **total_count**: `int` - Total count of watchlists found by the query.
    * **watchlists**: `List[WatchlistV2]` - The list `n` number of watchlists retrieved from the query, where `n=page_size`.
    """

    total_count: Optional[int] = Field(
        None,
        description="The total count of all watchlists.",
        examples=[10],
        alias="totalCount",
    )
    watchlists: Optional[List[WatchlistV2]] = Field(
        None, description="The list of watchlists."
    )



import sys
from csv import DictWriter
from datetime import datetime
from io import TextIOWrapper
from itertools import chain
from typing import Iterable
from typing import List
from typing import Type

from pydantic import BaseModel
from rich.console import Console
from rich.console import ConsoleRenderable
from rich.console import RichCast
from rich.table import Table

from _incydr_cli import console
from _incydr_sdk.utils import get_fields
from _incydr_sdk.utils import iter_model_formatted
from _incydr_sdk.utils import model_as_card


def date(dt: datetime):
    """render locale appropriate date"""
    return dt.strftime("%x")


def date_time(dt: datetime):
    """render locale appropriate date & time (w/ time truncated to minute precision)"""
    return dt.isoformat(sep=" ")[:16] + " UTC"


def measure_renderable(r: ConsoleRenderable):
    # the console.measure() method limits the max measurement size to the current size of the console
    # so to get the _actual_ size that a table would occupy with no wrapping, we need to make a large
    # console object
    return Console(width=100_000).measure(r).maximum


def models_as_table(
    model: Type[BaseModel],
    models: Iterable[BaseModel],
    columns: List[str] = None,
    title=None,
    flat=False,
):
    headers = list(get_fields(model, include=columns, flat=flat))
    tbl = Table(*headers, title=title, show_lines=True)
    for m in models:
        values = []
        for _name, value in iter_model_formatted(
            m, include=headers, flat=flat, render="table"
        ):
            if isinstance(value, BaseModel):
                value = model_as_card(value)
            elif not isinstance(value, (ConsoleRenderable, RichCast, str)):
                value = str(value)
            values.append(value)
        tbl.add_row(*values)
    return tbl


def table(
    model: Type[BaseModel],
    models: Iterable[BaseModel],
    columns: List[str] = None,
    title=None,
    flat=False,
):
    tbl = models_as_table(model, models, columns, title, flat)

    if not tbl.rows:
        console.print("No results found.")
        return
    with console.pager():
        # expand console and table so no values get truncated due to size of console, since we're using a pager
        console.width = tbl.width = measure_renderable(tbl)
        console.print(tbl, crop=False, soft_wrap=False, overflow="fold")


def csv(
    model: Type[BaseModel],
    models: Iterable[BaseModel],
    columns: List[str] = None,
    flat: bool = False,
    file: TextIOWrapper = sys.stdout,
):
    models = iter(models)
    try:
        first = next(models)
        models = chain([first], models)
    except StopIteration:
        console.print("No results found.")
        return
    headers = list(get_fields(model, columns, flat=flat))
    writer = DictWriter(file, fieldnames=headers, extrasaction="ignore")

    writer.writeheader()
    for m in models:
        flat_event = dict(
            iter_model_formatted(m, flat=flat, render="csv", include=headers)
        )
        writer.writerow(flat_event)



from __future__ import annotations

from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field

from _incydr_sdk.core.models import Model
from _incydr_sdk.core.models import ResponseModel


class UserTypes(str, Enum):
    USER = "USER"
    SUPPORT_USER = "SUPPORT_USER"
    API_CLIENT = "API_CLIENT"
    SYSTEM = "SYSTEM"
    UNKNOWN = "UNKNOWN"


class DateRange(BaseModel):
    endTime: Optional[float] = None
    startTime: Optional[float] = None


class AuditEventsPage(ResponseModel):
    """
    A model representing a page of audit events.

    **Fields**:

    * **events**: `List[Dict[Optional[str], Any]]` A list of zero or more events matching the given criteria.
        Each event is represented as a dictionary of property names associated with that event.  These fields may differ
        from event to event.
    * **pagination_range_end_index**: `int` The index of the last result returned, in relation to total results found.
    * **pagination_range_start_index**: `int` The index of the first result returned, in relation to total results found.
    """

    events: List[Dict[Optional[str], Any]] = Field(
        None, description="A list of zero or more events matching the given criteria."
    )
    pagination_range_end_index: int = Field(
        None,
        description="The index of the last result returned, in relation to total results found",
        examples=[62],
        alias="paginationRangeEndIndex",
    )
    pagination_range_start_index: int = Field(
        None,
        description="The index of the first result returned, in relation to total results found",
        examples=[0],
        alias="paginationRangeStartIndex",
    )


class QueryExportRequest(Model):
    actorIds: Optional[List[str]]
    actorIpAddresses: Optional[List[str]]
    actorNames: Optional[List[str]]
    dateRange: Optional[DateRange]
    eventTypes: Optional[List[str]]
    resourceIds: Optional[List[str]]
    userTypes: Optional[List[Union[UserTypes, str]]]


class QueryAuditLogRequest(Model):
    actorIds: Optional[List[str]]
    actorIpAddresses: Optional[List[str]]
    actorNames: Optional[List[str]]
    dateRange: Optional[DateRange]
    eventTypes: Optional[List[str]]
    page: int
    pageSize: int
    resourceIds: Optional[List[str]]
    userTypes: Optional[List[Union[UserTypes, str]]]



from typing import Optional

from pydantic import Field
from pydantic import model_validator

from _incydr_sdk.core.models import CSVModel
from _incydr_sdk.core.models import Model


class UserCSV(CSVModel):
    user: str = Field(csv_aliases=["user", "user_id", "username", "id", "userId"])


class UserJSON(Model):
    username: Optional[str] = None
    userId: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def _validate(cls, values):  # noqa
        if "username" not in values and "userId" not in values:
            raise ValueError("A json key of 'username' or 'userId' is required")
        return values

    @property
    def user(self):
        return self.userId or self.username


class AgentCSV(CSVModel):
    agent_id: str = Field(csv_aliases=["agentGuid", "agent_id", "agentId", "guid"])


class AgentJSON(Model):
    agent_id: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def _validate(cls, values):  # noqa
        if "agentGuid" in values:
            values["agent_id"] = values["agentGuid"]
            return values
        elif "agent_id" in values:
            return values
        elif "agentId" in values:
            values["agent_id"] = values["agentId"]
            return values
        elif "guid" in values:
            values["agent_id"] = values["guid"]
            return values
        else:
            raise ValueError(
                "A json key of 'agentGuid', 'agent_id', 'agentId', or 'guid' is required"
            )
